/*
 * net.c -- socket I/O -- This is a stripped down version of the
 * net.cc module from vic, with the following (original) copyright:
 *
 * Copyright (c) 1993-1994 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and the Network Research Group at
 *      Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include "net.h"
#include "utils.h"
#include "compat.h"
#include "rlcif.h"

/***
 ***
   We need the following routines:
   - open a multicast/unicast socket for send.
       (receiver: send request; sender: send data)
	openssock() does that;
   - open a multicast/unicast socket for receive
       (receiver: get request; sender: get data)
	openrsock() does that, it also join the
	m-cast group (if a multicast address is provided).
   - join a m-cast group
	join_group().
   - leave a m-cast group
	leave_group().
 ***
 ***/

static int
nonblock(int fd)
#ifdef DUMMYNET
{}
#else /* DUMMYNET */
{
#ifdef WIN32
    u_long flag = 1;

    if (ioctlsocket(fd, FIONBIO, &flag) == -1) {
	return (rlc_internal_set_error_ (-1, NET_CALL_ERROR,
	    "* ioctlsocket: FIONBIO ERROR"));
    }
#else 
    int flags = fcntl(fd, F_GETFL, 0);
#if defined(hpux) || defined(__hpux)
    flags |= O_NONBLOCK;
#else
    flags |= O_NONBLOCK|O_NDELAY;
#endif
    if (fcntl(fd, F_SETFL, flags) == -1) {
	return (rlc_internal_set_error_ (-1, NET_CALL_ERROR,
	    "* fcntl: F_SETFL ERROR"));
    }
#endif
    return 0;
}    
#endif /* DUMMYNET */

int
openssock(n32 addr, n16 port, int ttl, int unbound)
#ifdef DUMMYNET
{return(0);}
#else /* DUMMYNET */
{
    int fd;
    struct sockaddr_in sin;

    NDEB(fprintf(stderr,"--- openssock addr %s port %d\n",
	inet_ntoa(*(struct in_addr *)&addr), ntohs(port) ));

    fd = socket(AF_INET, SOCK_DGRAM, 0);
    if (fd < 0)
	return (rlc_internal_set_error_ (-1, NET_CALL_ERROR, "* socket call ERROR"));
    if (nonblock(fd) < 0)
	return (rlc_internal_set_error_ (-1, NET_CALL_ERROR, "* nonblock ERROR"));

    if (IN_CLASSD(ntohl(addr))) {
#ifdef IP_ADD_MEMBERSHIP
#ifdef WIN32
	u_int t;
#else
	u_char t;
#endif

	/*** set the multicast TTL ***/
	t = (ttl > 255) ? 255 : (ttl < 0) ? 0 : ttl;
	if (setsockopt(fd, IPPROTO_IP, IP_MULTICAST_TTL,
		   (char*)&t, sizeof(t)) < 0) {
	    return (rlc_internal_set_error_ (-1, NET_CALL_ERROR,
		"* setsockopt: IP_MULTICAST_TTL ERROR"));
	}
#else
	return (rlc_internal_set_error_ (-1, NET_CALL_ERROR,
		"* no multicast support ERROR"));
#endif
    }

    if (unbound)
	return fd;

#ifdef WIN32
    memset((char *)&sin, 0, sizeof(sin));
    sin.sin_family = AF_INET;
    sin.sin_port = 0;
    sin.sin_addr.s_addr = INADDR_ANY;
    if (bind(fd, (struct sockaddr *)&sin, sizeof(sin)) < 0) {
	return (rlc_internal_set_error_ (-1, NET_CALL_ERROR,
		"* bind call ERROR"));
    }       
#endif          
    memset((char *)&sin, 0, sizeof(sin));
    sin.sin_family = AF_INET;
    sin.sin_port = port;
    sin.sin_addr.s_addr = addr;
    if (connect(fd, (struct sockaddr *)&sin, sizeof(sin)) < 0) {
	return (rlc_internal_set_error_ (-1, NET_CALL_ERROR,
		"* connect call ERROR"));
    }

    return fd ;
}
#endif /* DUMMYNET */

int
join_group(int fd, n32 addr)
#ifdef DUMMYNET
{return(0);}
#else /* DUMMYNET */
{
  struct ip_mreq mr;
    /* if not m-cast addr. do nothing */
    if (!IN_CLASSD(ntohl(addr)))
	return(0);

    /*** XXX now we shuold check whether fd is a valif socket */

    mr.imr_multiaddr.s_addr = addr;
    mr.imr_interface.s_addr = INADDR_ANY;
    if (setsockopt(fd, IPPROTO_IP, IP_ADD_MEMBERSHIP,
	       (char *)&mr, sizeof(mr)) < 0) {
	return (rlc_internal_set_error_ (-1, NET_CALL_ERROR,
		"* setsockopt: IP_ADD_MEMBERSHIP ERROR"));
    }
    return (0);
}
#endif /* DUMMYNET */

int
leave_group(int fd, n32 addr)
#ifdef DUMMYNET
{return(0);}
#else /* DUMMYNET */
{
  struct ip_mreq mr;
    /* if not m-cast addr. do nothing */
    if (!IN_CLASSD(ntohl(addr)))
	return(0);

    /*** XXX now we shuold check whether fd is a valif socket */

    mr.imr_multiaddr.s_addr = addr;
    mr.imr_interface.s_addr = INADDR_ANY;
    if (setsockopt(fd, IPPROTO_IP, IP_DROP_MEMBERSHIP,
	       (char *)&mr, sizeof(mr)) < 0) {
	return (rlc_internal_set_error_ (-1, NET_CALL_ERROR,
		"* setsockopt: IP_DROP_MEMBERSHIP ERROR"));
    }
    return (0);
}
#endif /* DUMMYNET */

int
openrsock(n32 addr, n16 *port)
#ifdef DUMMYNET
{static n16 i=0; port = &i; return(0);}
#else /* DUMMYNET */
{
    int fd;
    struct sockaddr_in sin;
    int on = 1;
                
    NDEB(fprintf(stderr,"--- openrsock addr %s port %d\n",
	inet_ntoa(*(struct in_addr *)&addr), ntohs(*port) );)
    fd = socket(AF_INET, SOCK_DGRAM, 0);
    if (fd < 0) {
	return (rlc_internal_set_error_ (-1, NET_CALL_ERROR,
		"* socket call ERROR"));
    }
    if (nonblock(fd) < 0)
	return (rlc_internal_set_error_ (-1, NET_CALL_ERROR, "* nonblock ERROR"));
    if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR,(char *)&on, sizeof(on)) < 0) {
	return (rlc_internal_set_error_ (-1, NET_CALL_ERROR,
		"* setsockopt: SO_REUSEADDR ERROR"));
    }
#ifdef SO_REUSEPORT
    on = 1;
    if (setsockopt(fd, SOL_SOCKET, SO_REUSEPORT, (char *)&on, sizeof(on)) < 0) {
	return (rlc_internal_set_error_ (-1, NET_CALL_ERROR,
		"* setsockopt: SO_REUSEPORT ERROR"));
    }
#endif
    memset((char *)&sin, 0, sizeof(sin));
    sin.sin_family = AF_INET;
    sin.sin_port = *port;
#ifdef IP_ADD_MEMBERSHIP
    if (IN_CLASSD(ntohl(addr))) {

	/*
	 * Try to bind the multicast address as the socket
	 * dest address.  On many systems this won't work
	 * so fall back to a destination of INADDR_ANY if
	 * the first bind fails.
	 */
	sin.sin_addr.s_addr = addr;
	if (bind(fd, (struct sockaddr *)&sin, sizeof(sin)) < 0) {
	    rlc_have_error_ (-1, MULTI_BIND_ERROR, NULL);
	    sin.sin_addr.s_addr = INADDR_ANY;
	    if (bind(fd, (struct sockaddr*)&sin, sizeof(sin)) < 0) {
		return (rlc_internal_set_error_ (-1, NET_CALL_ERROR,
		    "* bind call ERROR"));
	    }
	}
	if (join_group(fd, addr)!=0)
	    return (fd);
    } else
#endif
    {
	/*
	 * bind the local port to this socket.  If that
	 * fails, another app probably has the addresses bound so
	 * just exit.
	 */
	sin.sin_addr.s_addr = addr ;
	if (bind(fd, (struct sockaddr *)&sin, sizeof(sin)) < 0) {
	    return (rlc_internal_set_error_ (-1, NET_CALL_ERROR,
		"* bind call ERROR"));
	}
    }

    if (*port == 0) {
	memset((char *)&sin, 0, sizeof(sin));
	on = sizeof(sin);
        getsockname(fd, (struct sockaddr *)&sin, &on) ;
	*port = sin.sin_port ;
    }
    NDEB(fprintf(stderr, "--- port bound to %s : %d\n",
	inet_ntoa(sin.sin_addr), ntohs(*port) );)
    return fd ;
}
#endif /* DUMMYNET */

unsigned long
localaddr(int ssock_)
#ifdef DUMMYNET
{return 0;}
#else /* DUMMYNET */
{
  struct sockaddr_in ad;
  struct sockaddr_in *p = &ad;
    int len = sizeof(*p);   
    memset((char *)p, 0, sizeof(*p));
    p->sin_family = AF_INET;
    if (getsockname(ssock_, (struct sockaddr *)p, &len) < 0) {
	/* perror("getsockname");  */
	memset((char *)p, 0, sizeof(*p));
    }
    if (p->sin_addr.s_addr == 0) { 
	char hostname[80];
	struct hostent *hp;

	if (gethostname(hostname, sizeof(hostname)) >= 0) {
	    if ((hp = gethostbyname(hostname)) >= 0) {
		p->sin_addr.s_addr = ((struct in_addr *)hp->h_addr)->s_addr;
	    }
	}
    }
    NDEB(fprintf(stderr, "localname addr is %s\n",
	inet_ntoa(p->sin_addr) );)
    if (p->sin_addr.s_addr != 0)
	return (p->sin_addr.s_addr);
    else
	return -1;
}
#endif /* DUMMYNET */
