/*
 * receiver.c -- receiver routines for rlc.
 *
 * This file is part of
 * 
 * rlc -- Reliable Multicast data Distribution Protocol
 * 
 * (C) 1996, 1997, 1998 Luigi Rizzo and Lorenzo Vicisano
 *     (luigi@iet.unipi.it, vicisano@cs.ucl.ac.uk)
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Luigi Rizzo,
 *      Lorenzo Vicisano and other contributors.
 * 4. Neither the name of the Authors nor the names of other contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 * IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * receiver main routines
 */

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <fcntl.h>
#include <assert.h>
#include "rlc.h"
#include "rlcif.h"
#include "agent.h"
#include "net.h"
#include "congestion.h"
#include "layers.h"
#include "ui.h"
#include "utils.h"
#include "init.h"
#include "event.h"
#include "receiver.h"

				/*+ the event loop desctiptor is unique
				   for sender and receiver, and is in
				   init.c (rlc_events). +*/

				/*+ this is the pointer to the function
				   used to add/remove IO descriptor to
				   the client IO list, when the library
				   is used in client driven mode (without
				   looping on rlc_getevent */
extern rlc_app_chanhandle_t client_chanhandle;

				/*+ globally associates layers to sockets +*/
static int chans[MAX_RX_SOCKS];

static int receiver_init = 0;

static int get_data(void *p);

/**
 ** We have been woken up by the client, it should mean that
 ** we have received some data, so first check and then call
 ** get_data. Note that checking we also make `rlc_showselected'
 ** return the proper value .
 **/
static void ex_wakeup()
{
  int i;
  void *p;
    if (rlc_poolchan(&i, &p) >= 0 &&
	i == ID_GETDATA &&
	p )
	(void)get_data(p);
}

/** 
 ** various auxiliary fonctions **
 **/

#define	NO_SOCK	-1
/*
 * Join group, possibly creating socket.
 */ 
int do_join (int pi, int g)
{
  ui16 tport;
  sxs_t *sxs = (sxs_t *)_rlc_getinstance(pi, rlc_inst_mngr);
    if (!sxs)
	return (rlc_internal_set_error_ (pi, JOIN_ERROR, NULL));

    tport = htons(ntohs(sxs->port) + 2 * g);
    assert(g>=0 && g<MAX_LAYERS);
    if (sxs->socks[g] == NO_SOCK) {
	if ((sxs->socks[g] = openrsock(sxs->addrss[g], &tport)) < 0)
	    return (rlc_internal_set_error_ (pi, JOIN_ERROR, NULL));
	if (fcntl (sxs->socks[g], F_SETFL, O_NONBLOCK) < 0) {
	    return (rlc_internal_set_error_ (pi, JOIN_ERROR, NULL));
	}
	if (sxs->socks[g] < 0) {
	    sxs->socks[g] = NO_SOCK;
	    return (rlc_internal_set_error_ (pi, JOIN_ERROR, NULL));
	}
    } else {
	struct sockaddr_in srcaddr;
	char buff_in[MAX_PKT_LEN];
	int srcaddrlen = sizeof(srcaddr);
	/* discard old packets...
	   but shouldn't thrown away! (but there shouldn't be!) XXX */
	while ( recvfrom(sxs->socks[g], buff_in, sizeof(buff_in), 0,
	    (struct sockaddr *)&srcaddr, &srcaddrlen) > 0) ;
	if (join_group(sxs->socks[g], sxs->addrss[g]) != 0)
	    return (rlc_internal_set_error_ (pi, JOIN_ERROR, NULL));
		
    }
    NDEB(fprintf(stderr, "group %x  joined (port %d)\n", sxs->addrss[g],
	tport);)
    if (sxs->socks[g] >= MAX_RX_SOCKS) {
	return (rlc_internal_set_error_ (pi, TOO_MANY_SOCKET, NULL));
    }
    chans[sxs->socks[g]] = g;
    if (rlc_insertchan(sxs->socks[g], ID_GETDATA, get_data,
	 (void *)sxs) < 0) {
	return (rlc_internal_set_error_ (pi, NO_CHAN, NULL));
    }
    if (client_chanhandle)
	(*client_chanhandle)(sxs->socks[g], &ex_wakeup);

#ifdef HAVE_UI
    if (sxs->have_gui)
	rlc_gui_tcl_exec (pi, NULL, "set_current %d", g);
#endif /* HAVE_UI */

    return (0);
}

int do_leave (int pi, int g)
{
  sxs_t *sxs = (sxs_t *)_rlc_getinstance(pi, rlc_inst_mngr);
    if (!sxs) return(-1);

    assert(g>=0 && g<MAX_LAYERS);
    if (sxs->socks[g] == NO_SOCK)
	return (0);
    if (leave_group(sxs->socks[g], sxs->addrss[g]) != 0)
	return (rlc_internal_set_error_ (pi, LEAVE_ERROR, NULL));

#ifdef HAVE_UI
    if (sxs->have_gui)
	rlc_gui_tcl_exec (pi, NULL, "set_current %d", g-1);
#endif /* HAVE_UI */
    DEB (fprintf(stderr, "do_leave: %d left\n", g);)

    /* note that we don't call rlc_deletechan to allow to suck packets anyway!
       this call will be made when the protocol instance terminates */
     
    return(0);
}

void
receiver_last_stuff(int i)
{
  int j;
  ui32 endtime;
  sxs_t *sxs;
  int id = GET_SESSION_START;

    while ((id = rlc_get_allinst(id, rlc_inst_mngr))
	    != GET_SESSION_END ) {
        sxs = (sxs_t *)_rlc_getinstance(id, rlc_inst_mngr);
	if (!sxs->isreceiver) continue;
	if (!sxs->is_on) { /* ..... XXX */
	    DEB(fprintf(stderr, "rlc: last_stuff for %d (%x)\n",
		sxs->ides, (int)sxs);)
	    endtime = timestamp_u();
	    /* the following leaves all groups as well */
	    congestion_end(sxs->ides, sxs->cong, endtime);
	    for (j=0; sxs->socks[j] != NO_SOCK && j<MAX_LAYERS; j++) {
		rlc_deletechan(sxs->socks[j]);
		if (client_chanhandle)
		    (*client_chanhandle)(sxs->socks[j], NULL);
		close(sxs->socks[j]);
	    }
	    NDEB(printf("Stats: \n"
		"in %d sec. ???\n",
		(int)((endtime-sxs->starttime)/1e6));)
	    _rlc_return_instance((instance_des_t *) sxs, rlc_inst_mngr);
	    free(sxs);
	}
    }
    if (i == SIGINT ) exit(i);
}

cong_t *
get_cong_inst(int pi)
{
  sxs_t *sxs = (sxs_t *)_rlc_getinstance(pi, rlc_inst_mngr);
    if (!sxs) return(NULL);

    return(sxs->cong);
}

/* that's to terminate asynchronously the reception */
static int
force (int pi)
{
  sxs_t *sxs = (sxs_t *)_rlc_getinstance(pi, rlc_inst_mngr);
    DEB(fprintf(stderr, "rlc force for %d (%x)\n", pi, (int)sxs);)
    if (!sxs) return(-1);
    sxs->is_on = 0;
    if (!sxs->termdelayed) {
	DEB(fprintf(stderr, "rlc force no delay: terminating\n");)
	if (sxs->watch_t && _rlc_howmany_instance(rlc_inst_mngr) == 1) {
	    DEB(fprintf(stderr, "rlc force deleting RLC_WATCHDOG\n");)
	    rlc_deletetimer(RLC_WATCHDOG);
	}
	receiver_last_stuff(0);
    }
    return(0);
}

#define APPL_SPEC	-3
static int 
chk_pkt(rlc_h *header, sxs_t *sxs, int chan, int src)
{
  int this_lay;

    assert(chan < MAX_RX_SOCKS);
    this_lay = chans[chan];
    assert(sxs->socks[this_lay] == chan);

    if (header->magic != htons(RLC_MAGIC)) {
NDEB(fprintf(stderr, "chk_pkt: incorrect magic number\n"));
	return(-1);
    }

    if (!sxs->src)		/* not bound */
	sxs->src = src;
    else if (sxs->src != src)	/* bound!! */
	return -1;

    if (header->vers != 1) {
NDEB(fprintf(stderr, "chk_pkt: only understand version 1, this is %d\n",
	    header->vers));
	return(-1);
    }
    if (header->type != RLC_DATA_T) {
	if (header->type != RLC_APPL_T)
	    return(APPL_SPEC);
	/* this also includes RLC_WRMP_T */
NDEB(fprintf(stderr, "chk_pkt: incorrect data packet, discarding\n"));
	return(-1);
    }
    if (header->bp != BT_EXP_1) {
	/* in this implementation we don't support this ! */
NDEB(fprintf(stderr, "chk_pkt: we only implement band. profile"
		" BT_EXP_1!\n"));
	rlc_have_error_ (sxs->ides, BAND_NOT_IMPL, NULL);
	return -1;
    }
    if ((header->flags & RLC_F_EDGE_TRIG) != 0) {
	/* in this implementation we don't support this ! */
NDEB(fprintf(stderr, "chk_pkt: RLC_F_EDGE_TRIG is not stucked at 0!\n"));
	rlc_have_error_ (sxs->ides, TRIG_NOT_IMPL, NULL);
	return -1;
    }
    if (!sxs->have_addr_ass) {
	if (!(header->flags & RLC_F_CONS_ADDR)) {
	NDEB(fprintf(stderr, "chk_pkt: waiting for address assignement"
			", discarding\n"));
		return(-1);
	}
    } else if (header->flags & RLC_F_CONS_ADDR) {
	sxs->have_addr_ass = 0;
	/* and do something else */
    }

    return(this_lay);
}

static inline void update_guess(int *lkypays, int *lkypaysv, int thispays)
{
    if (*lkypays > thispays) {
	*lkypaysv = (*lkypaysv + *lkypays - thispays) >> 1;
    } else {
	*lkypaysv = (*lkypaysv + thispays - *lkypays) >> 1;
    }
    *lkypays = (*lkypays + thispays) >> 1;
   NDEB(fprintf(stderr, "***update_guess now %u %u\n",
	*lkypays, *lkypaysv);)
}

 /*
 ** this is going to become pretty complex:
 ** `somehow' we try to guess the most likely payload size (lkypays)
 ** and its sort-of-variance (lkypaysv), then we read a scattered array like
 ** that: [[sizeof(header)] [lkypays+lkypaysv] [sizeof(safety_buff)]].
 **
 */
static char *
read_pkt(sxs_t *sxs, int chan, rlc_h **header, int *len, ui32 *src,
    int *lkypays, int *lkypaysv)
{
  char *buff;
  struct iovec iov[3];
  struct msghdr msg;
  struct sockaddr_in srcaddr;
  char safety_buff[MAX_PAYLOAD_SIZE];
  int doublev = *lkypaysv << 1;


    if ((*header = (rlc_h *)malloc(sizeof(rlc_h))) == NULL) {
	rlc_have_error_ (sxs->ides, ALLOC_PKT_ERROR, NULL);
	return NULL;
    }

    if ((buff = malloc(*lkypays+doublev)) == NULL) {
	rlc_have_error_ (sxs->ides, ALLOC_PKT_ERROR, NULL);
	return NULL;
    }

    iov[0].iov_base = (caddr_t)*header;
    iov[0].iov_len = sizeof(rlc_h);
    /* XXX what about if we have header extensions ??? */
    iov[1].iov_base = (caddr_t)buff;
    iov[1].iov_len = *lkypays+doublev;
    iov[2].iov_base = (caddr_t)safety_buff;
    iov[2].iov_len = sizeof(safety_buff);
    bzero((char *) &msg, sizeof(msg));
    msg.msg_iov = iov;
    msg.msg_iovlen = 3;
    msg.msg_name = (caddr_t)&srcaddr;
    msg.msg_namelen = sizeof(srcaddr);
    if ((*len = recvmsg(chan, &msg, 0)) <=0) {
	rlc_have_error_(sxs->ides, RCV_ERROR, NULL);
	free(buff);
	return(NULL);
    }

    /* have we failed our guess ? */
    if (*len > sizeof(rlc_h)+*lkypays+doublev) {
	void *pt;
	if ((pt = realloc(buff, *len-sizeof(rlc_h))) == NULL) {
	    rlc_have_error_ (sxs->ides, ALLOC_PKT_ERROR, NULL);
	    return NULL;
	}
	buff = pt;
	bcopy(safety_buff, buff+*lkypays+doublev,
	    *len-(sizeof(rlc_h)+*lkypays+doublev));
NDEB(fprintf(stderr, "read_pkt: Guess wrong (read %d)!!\n",
	    *len-sizeof(rlc_h));)
    }

    update_guess(lkypays, lkypaysv, *len-sizeof(rlc_h));

   NDEB(fprintf(stderr,
	"--- read_pkt: pkt %u sess %ld from: %lx, siz %d\n",
	ntohs((*header)->lsn), ntohl((*header)->suid),
	(unsigned long)srcaddr.sin_addr.s_addr, *len));

    *src = (unsigned long)srcaddr.sin_addr.s_addr;

    return(buff);
}

static pkt_lis_d *pkt_lis_h;		/*+ this is a pointer to a temporary
					    queue of received packets: all
					    the received packet gets read from
					    the sockets and reordered before
					    delivering them to the application
					    and running the congestion control
					    algorithm +*/
static int
get_data(void *p)
{
  sxs_t *sxs;
  int len, chan, this_lay;
  char *buff_in;
  ui32 this_time;
  int paylen;
  rlc_h *header;
  pkt_lis_d **plp, *lp;
  int i;
  ui32 tidx;
  ui32 src;

    do {
	sxs = (sxs_t *)p;
	chan = rlc_showselected();

	if (NULL == (buff_in = read_pkt(sxs, chan, &header, &len,
		&src, &(sxs->lkypays), &(sxs->lkypaysv)))) {
	    continue;
	}

	assert((long)len - (long)sizeof(*header) >= 0 &&
	    len - sizeof(*header) < MAX_PAYLOAD_SIZE);

	if ((this_lay = chk_pkt(header, sxs, chan, src)) < 0) {
	    if (sxs->rx_packet != NULL && this_lay == APPL_SPEC)
		sxs->rx_packet(sxs->cpi, 0,
		    '\1' /* XXX */, buff_in, len - sizeof(*header));
	    else	
		free(buff_in);
	    continue;
	}

	/* this is for the watchdog */
	sxs->watch_last = timestamp_u();

	/* and insert in queue... have to save:
	   p, buff_in, header, len, chan?  */
	tidx = global_index (ntohs(header->lsn), ntohs(header->pn),
		this_lay, (int)header->layer_n - 1,
		is_in_burst(header->flags));
	lp = NULL;
	if (!pkt_lis_h) plp = &pkt_lis_h;
	else
	    for (plp = &(pkt_lis_h->next); *plp != NULL; plp = &((*plp)->next))
		if (ui32diff(tidx, (*plp)->gindex) < 0) {
		    lp = *plp;
		    break;
		}
	if ( (*plp = malloc(sizeof(pkt_lis_d))) == NULL) {
	    rlc_have_error_ (sxs->ides, ALLOC_TMP_ERROR, NULL);
	    return 0;
	}
	(*plp)->p = p; (*plp)->buff_in = buff_in;
	(*plp)->header = header; (*plp)->len = len;
	(*plp)->next = lp; (*plp)->gindex = tidx;
	(*plp)->this_lay = this_lay;

    } while (
	rlc_poolchan(&i, &p) >= 0 &&
	i == ID_GETDATA &&
	p );

    sxs->termdelayed = 1;
    for (i = 0; pkt_lis_h != NULL && sxs->is_on;
	    lp = pkt_lis_h, pkt_lis_h = pkt_lis_h->next, free(lp), i++) {
	p = pkt_lis_h->p; buff_in = pkt_lis_h->buff_in;
	header = pkt_lis_h->header; len = pkt_lis_h->len;
	this_lay = pkt_lis_h->this_lay; tidx = pkt_lis_h->gindex;

	sxs = (sxs_t *)p;
	paylen = len - sizeof(*header);

	/* XXX this should be saved before... */
	this_time = timestamp_u();

	/* perform congestion control */
	congestion_control(sxs->ides, sxs->cong, header, paylen,
		this_lay, tidx, this_time);

	/* this must be the last .... */
	if (sxs->rx_packet != NULL && !(header->flags & RLC_F_HART_BEAT))
	    sxs->rx_packet(sxs->cpi, this_lay,
	    '\0' /* XXX */, buff_in, paylen);

    }
    sxs->termdelayed = 0;
    if (!sxs->is_on)
	force (sxs->ides);
	
    return(0);
}

void
_rlc_rx_init()
{
    if (!receiver_init) {
	pkt_lis_h = NULL;
	receiver_init = 1;
    }
}


/*++++
 
 This registers a function that gets called when an IO descriptor
 must be added/deleted from the list of descriptor to be polled
 (select). This function is intended only for use in the client
 driven mode (the one where the client controls the event loop),
 and must be called before any other call to the library.

 Params:
	- f : pointer to the function to be called by the library. The
	      argument function is prototyped as :
			typedef void (*rlc_app_chanhandle_t)
				(int iodes, void (f)());
	      where iodes is the IO descriptor to be observed, and f
	      is the function to be called whe the descriptor is ready for
	      reading. If f = NULL the descriptor must be removed.
	      This function is implemented by the client.

 Returns:
	0 on succes, < 0 on failure.
*/
int
rlc_rx_registerchanhandle(rlc_app_chanhandle_t f)
{
    if (f && !rlc_events) {
	client_chanhandle = f;
	return(0);
    }
    return(-1);
}

/*++++
 
 This sets up a watchdog timer to detect reception stall. It
 should be used in rlc driven mode only. If a stall is
 detected, rlc either call the provided callback function
 or rise an asynhronous error (RECEPTION_TIMEOUT_WAR).

 Params:
	pi : protocol instance ID.

	f : pointer to the function to be called by rlc. If a NULL
		argument is given, the timeout procedure will rise
		an error condition.  The argument function is
		prototyped in `rlcif.h' as :
			typedef void (*rx_watchdog_t) (int pi);
		pi is the rlc protocol instance descriptor.
	time : the max idle time, if = 0 the default of 30s is used.

 Returns:
	0 on success, < 0 on failure.
*/
int
rlc_rx_set_watchdog(int pi, rx_watchdog_t f, int time)
{
  sxs_t *sxs = (sxs_t *)_rlc_getinstance(pi, rlc_inst_mngr);
    if (!sxs)
	return (rlc_return_error_ (pi, CANT_SET_WATCHDOG, NULL));

    sxs->watch_f = f;
    if (time > 0)
	sxs->watch_t = time;
    else
	sxs->watch_t = 30e6;	/* default 30s */

    return(0);
}

int
rlc_watchdog_(void *pi)
{
  sxs_t *sxs = (sxs_t *)_rlc_getinstance((int)pi, rlc_inst_mngr);
  ui32 now=timestamp_u();
    if (!sxs)
	return 0;

   NDEB(fprintf(stderr, "rlc_watchdog_ now %d, last %d, t %d\n",
	now, sxs->watch_last, sxs->watch_t);)

    if ((int)(now - sxs->watch_last - sxs->watch_t) >= 0) {
	/** hops, we have been too much idle! */
	if (sxs->watch_f)
	    (*sxs->watch_f)((int)pi);
	else
	    rlc_have_error_ ((int)pi, RECEPTION_TIMEOUT_WAR, NULL);
    }

    if (sxs->is_on && sxs->watch_t && rlc_settimer(RLC_WATCHDOG, sxs->watch_t,
	    (CALLBACK)rlc_watchdog_, (void *)pi, 0) < 0) {
	rlc_have_error_ ((int)pi, CANT_SET_WATCHDOG, NULL);
    }

    return 0;
}


/*++++
 
 Starts the reception by subscribing the first layer.
 Each time a packet is available rx_packet gets called to consume
 the packet. The reception is terminate by explicitly calling
 `rlc_abort' for this protocol instance.
    
 Params:
	pi : protocol instance ID.

 Returns:
	0 on success, < 0 on failure, > 0 if a warning
	is generated, use rlc_error_info() to get error
	or warning description. Possible error/warning
	causes are:
		- CANT_START_RECEIVER (cannot start this receiver).
*/
int
_rlc_rx_start(int pi)
{
  int res;
  sxs_t *sxs = (sxs_t *)_rlc_getinstance(pi, rlc_inst_mngr);
    if (!sxs)
	return (rlc_return_error_ (pi, NO_SUCH_INSTANCE, NULL));

    /* XXX specific receiver checks ... there should be none */

    if ((res = do_join(pi, 0)) < 0)
	return (rlc_return_error_ (pi, res, NULL));

    if (sxs->watch_t && !client_chanhandle &&
		 (res = rlc_settimer(RLC_WATCHDOG, sxs->watch_t,
		 (CALLBACK)rlc_watchdog_, (void *)pi, 0)) < 0) {
	return (rlc_return_error_ (pi, res, "rlc_settimer error"));
    }

    sxs->watch_last = timestamp_u();
    sxs->src = 0;			/* not bounded */
    return(0);
}

/*++++
 
 Terminates the reception unsubscribing all the currently joined layers.
 It also frees the instance descriptor.
    
 Params:
	pi : protocol instance ID.

 Returns:
	0 on success, < 0 on failure.
*/

int
_rlc_rx_abort(int pi)
{
    return(force(pi));
}

/*++++

 Retrieves the source address.

 Params:
	pi : protocol instance ID.
 
 Returns:
	the source address on success, -1 on failure.
*/
unsigned long
rlc_rx_get_src (int pi)
{
  sxs_t *sxs = (sxs_t *)_rlc_getinstance(pi, rlc_inst_mngr);
    if (!sxs) {
	rlc_internal_set_error_ (pi, NO_SUCH_INSTANCE, NULL);
	return (-1);
    }
    if (sxs->src == 0)
	return (rlc_return_error_  (pi, -1, "No source address available"));

    return (sxs->src);
}

