/*
 * rlc.c -- sender/receiver setup routines
 *
 * This file is part of
 * 
 * rlc -- Reliable Multicast data Distribution Protocol
 * 
 * (C) 1996, 1997, 1998 Luigi Rizzo and Lorenzo Vicisano
 *     (luigi@iet.unipi.it, vicisano@cs.ucl.ac.uk)
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Luigi Rizzo,
 *      Lorenzo Vicisano and other contributors.
 * 4. Neither the name of the Authors nor the names of other contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 * IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>	/* for stat: without this doesn't
			   compile in FreeBSD 2.1 */
#include <netinet/in.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <sys/stat.h>	/* for stat */
#include <unistd.h>	/* for sleep */
#if (defined(Linux))
	#include <sys/uio.h>
#endif
#if (!defined(Linux))
#if (defined(FreeBSD) || (defined(SunOS) && (OSMVER < 5)))
	#include <errno.h>
	#include <sys/uio.h>
#endif
#endif
#include "rlc.h"
#include "rlcif.h"
#include "agent.h"
#include "ui.h"
#include "sender.h"
#include "receiver.h"
#include "layers.h"
#include "utils.h"
#include "compat.h"
#include "init.h"
#include "event.h"
#include "error.h"

rlc_app_chanhandle_t client_chanhandle = NULL;

/*++++

Creates a new protocol instance assigning a local protocol instance
identifier.

Returns:
    - the protocol instance id, on success.
    - a negative value on failure:

*/

int
rlc_new(int isreceiver)
{
  int ret, retvalue, i;
  sxs_t *se;

    /* library initialization ? */
    if ((ret = rlc_internal_init()) != 0) {
	if (ret > 0)
	    retvalue = ret;
	else
	    return (rlc_return_error_ (-1, ret, "Error in rlc_internal_init()"));
    }
    _rlc_rx_init();
    _rlc_tx_init();

    /**
     ** allocate session descriptor
     **/
    if (NULL == (se = calloc(1, sizeof(sxs_t)))) {
	return (rlc_return_error_ (-1, ALLOC_ERROR,
	    "* Cannot allocate session descriptor"));
    }

    if (NULL == (se->cong = calloc(1, sizeof(cong_t)))) {
	free (se);
	return (rlc_return_error_ (-1, ALLOC_ERROR, NULL));
    }

    if (_rlc_insert_instance ((instance_des_t *) se, rlc_inst_mngr) < 0) {
	free(se);
	return (rlc_return_error_(-1, TOO_MANY_INSTANCES, NULL));
    }

    se->is_on = 0;			/* start will set to true */
    se->port = 0;
    se->cpi = 0;
    se->have_gui = 0;
#ifdef HAVE_UI
    se->error_gui = NULL;
    se->warning_gui = NULL;
#endif
    se->geometry = NULL;
    se->starttime = 0;			/* will be set by start */
    se->isreceiver = isreceiver;

    /* Sender Specific */
    se->ch = 0;
    se->ttl = TTL;
    se->rate = L0_PERIOD;
    se->mpkts = DATA_SIZE;
    se->Mpkts = DATA_SIZE;
    se->pkts_left = INFINITE;		/* this is not used ! */
    se->tx_packet = NULL;
    se->flags = 0;
    se->l.nc = NLAY;

    /* Receiver Specific */
    se->termdelayed = 0;
    se->src = 0;
    se->lkypays = MAX_PAYLOAD_SIZE;
    se->lkypaysv = 0;
    se->rx_packet = NULL;
    assert(congestion_setup(se->ides, se->cong) >= 0);
    for (i=0; i<MAX_LAYERS; i++)
	se->socks[i] = NO_SOCK;
    se->addrss[0] = -1;			/* this to detect if addresses
					 * has ben set */
    se->have_addr_ass = 0;
    /* se->addrss ? */
    se->watch_t = 0;
    se->watch_last = 0;
    se->watch_f = NULL;

    /* XXX the wornings get lost !!! */
    return (se->ides);
}

/*++++

Provides the multicast address(es), port and ttl. ttl is ingnored
in the receiver.

Returns:
    - 0 on success.
    - a negative value on failure:
    - > 0 in caso of warning:

*/
int
rlc_set_addr(int pi, addr_list_t *rlcaddr)
{
  int consecutive;
  sxs_t *sxs = (sxs_t *)_rlc_getinstance(pi, rlc_inst_mngr);

    if (!sxs)
        return (rlc_return_error_ (pi, NO_SUCH_INSTANCE, NULL));

    if (sxs->is_on)
	return (rlc_return_error_ (pi, CANT_CHANGE_ADDR, NULL));

    /**
     ** start checking parameters
     **/

    /* port */
    sxs->port = htons(rlcaddr->port);
    if (ntohs(sxs->port)%2) {
	sxs->port = htons(ntohs(sxs->port)-1);
	NDEB(fprintf(stderr, "Warning: rlc_tx_setup called with odd port #,"
		" now port # = %u\n", ntohs(sxs->port));)
    }
    rlcaddr->port = ntohs(sxs->port);

    /* ttl */
    sxs->ttl = max (rlcaddr->ttl, 0);
    sxs->ttl = min (sxs->ttl, 255);
    rlcaddr->ttl = sxs->ttl;
    NDEB(fprintf(stderr, "rlc_tx_setup: have ttl %d\n", sxs->ttl);)

    /* addrs */
    hton_addlist(rlcaddr);
    if ((consecutive = set_addresses(rlcaddr, MAX_LAYERS, sxs->addrss)) < 0) {
	return (rlc_return_error_ (pi, consecutive, NULL));
    }

    if (consecutive) {
	sxs->have_addr_ass = 0;
	sxs->flags |= ADDR_CONS;
    } else {
	sxs->have_addr_ass = 1;
	sxs->flags &= (~ADDR_CONS);
    }

    return (0);
}

/*++++

Provides a reference to the client (cpi) and a pointer to
a callback function used to get/provide the packets being
transmitted/received.

Returns:
    - 0 on success.
    - a negative value on failure:
    - > 0 in caso of warning:

*/

int
rlc_attach(int pi, int cpi, void *f)
{
  sxs_t *sxs = (sxs_t *)_rlc_getinstance(pi, rlc_inst_mngr);

    if (!sxs)
        return (rlc_return_error_ (pi, NO_SUCH_INSTANCE, NULL));

    sxs->cpi = cpi;
    
    /* check tx_packet */
    if (f == NULL)
	return (rlc_return_error_(pi, BAD_CALLBACK_FUNCTION, NULL));
    sxs->tx_packet = (rlc_tx_packet_t)f;
    sxs->rx_packet = (rlc_rx_packet_t)f;

    return(0);
}

int
rlc_gui_enable(int pi)
{
  int retvalue = 0;
  sxs_t *sxs = (sxs_t *)_rlc_getinstance(pi, rlc_inst_mngr);
    if (!sxs)
	return (rlc_return_error_ (pi, NO_SUCH_INSTANCE, NULL));

#ifdef HAVE_UI
    if (!client_chanhandle) {
	if (rlc_gui_init(pi, 0, "rlc", sxs->geometry) < 0) {
	    retvalue = NOUI_WARNING;
	} else {
	    if (sxs->isreceiver && rlc_gui_rx_init (pi) < 0)
		retvalue = NOUI_WARNING;
	    if (!sxs->isreceiver && rlc_gui_tx_init (pi) < 0)
		retvalue = NOUI_WARNING;
	    if (_rlc_error_gui_init (pi) < 0)
		retvalue = NOUI_WARNING;
	}
    }

#else	/* HAVE_UI */
    retvalue = NOUI_WARNING;
#endif	/* HAVE_UI */
    if (!retvalue)
	sxs->have_gui = 1;
    return retvalue;
}
    
int
rlc_gui_geometry(int pi, char *geometry)
{
  sxs_t *sxs = (sxs_t *)_rlc_getinstance(pi, rlc_inst_mngr);

    if (!sxs)
	return (rlc_return_error_ (pi, NO_SUCH_INSTANCE, NULL));

#ifdef HAVE_UI
    if (geometry) {
	sxs->geometry = malloc (strlen(geometry)+1);
	strcpy (sxs->geometry, geometry);
    }
#else
    return NOUI_WARNING;
#endif
    return 0;
}

/*++++
 
 This registers a function that gets called when an IO descriptor
 must be added/deleted from the list of descriptor to be polled
 (select). This function is intended only for use in the client
 driven mode (the one where the client controls the event loop),
 and must be called before any other call to the library.

 */

int
rlc_registerchanhandle(rlc_app_chanhandle_t f)
{
    if (f && !rlc_events) {
	client_chanhandle = f;
	return(0);
    }
    return(-1);
}

/*++++
 
 Starts the reception/transmission.

 In the reciever starts the reception by subscribing the first layer.
 Each time a packet is available rx_packet gets called to consume
 the packet. The reception is terminate by explicitly calling
 `rlc_abort' for this protocol instance.
    
Returns:
    - 0 on success.
    - a negative value on failure:
    - > 0 in caso of warning:

 */

int
rlc_start(int pi)
{
  sxs_t *sxs = (sxs_t *)_rlc_getinstance(pi, rlc_inst_mngr);

    if (!sxs)
	return (rlc_return_error_ (pi, NO_SUCH_INSTANCE, NULL));
    /* XXX add all the checkes ... now seems OK */
    if (sxs->addrss[0] == -1)
	return (rlc_return_error_ (pi, BAD_ADDRESS, NULL));
    /* XXX check the client_chanhandle stuff */
    if (client_chanhandle == NULL &&
	sxs->tx_packet == NULL && sxs->rx_packet == NULL)
	return (rlc_return_error_ (pi, BAD_CALLBACK_FUNCTION, NULL));

    sxs->starttime = timestamp_u();
    sxs->is_on = 1;

    if (sxs->isreceiver)
	return (_rlc_rx_start(pi));
    else
	return (_rlc_tx_start(pi));
}

/*++++
 
 Stops the reception/transmission.

Returns:
    - 0 on success.
    - a negative value on failure:
 */
int
rlc_abort(int pi)
{
  sxs_t *sxs = (sxs_t *)_rlc_getinstance(pi, rlc_inst_mngr);

    if (!sxs)
	return (rlc_return_error_ (pi, NO_SUCH_INSTANCE, NULL));

    if (sxs->isreceiver)
	return (_rlc_rx_abort(pi));
    else
	sxs->is_on = 0;

    return 0;
}
