/*
 * utils.c --  some utility functions.
 *
 * This file is part of
 * 
 * rlc -- Reliable Multicast data Distribution Protocol
 * 
 * (C) 1996, 1997, 1998 Luigi Rizzo and Lorenzo Vicisano
 *     (luigi@iet.unipi.it, vicisano@cs.ucl.ac.uk)
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Luigi Rizzo,
 *      Lorenzo Vicisano and other contributors.
 * 4. Neither the name of the Authors nor the names of other contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 * IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#if (!defined(Linux))
#if ( defined(SunOS) && (OSMVER < 5))
#include <sys/types.h>		/* looks like that netinet/in.h need it */
#endif
#endif
#include<netinet/in.h>
#include "compat.h"
#include "utils.h"
#include "rlcif.h"
#include "event.h"


/**
 ** compute hopefully unique id
 **/
ui32 compute_id()
{
  short seed[3];
  struct timeval getime;

    gettimeofday(&getime, NULL);
    seed[0] = (short)getime.tv_usec; seed[1] = (short)(getime.tv_usec*0xffff);
    seed[2] = (short)getime.tv_sec;
    return(nrand48(seed));      
}

/*** parameter parsing ***/
int     
set_link_addr(char *s, char *a, int addrlen, n16 *p, char *msg)
{       
    char *s1, *q = a ;
    strncpy(q, s, addrlen);
    s1=strchr(q,'/');
    if (s1) {
        *p=htons(atoi(s1+1));
        *s1='\0'; 
	NDEB(printf("%s session [%s] [%d]\n", msg, q, ntohs(*p));)
    } else  
        return -1;
    return 0;
}

/*
** return values: < 0 on failure, >= 0 on success. If == 1 address are
** consecutive.
** Assume addresses are in net order.
*/
int
set_addresses(addr_list_t *list, int lays, ui32 *dstaddr)
{
  int consecutive, i;
  ui32 prev;
    if (list->nadd < 1)
	return (rlc_internal_set_error_(-1, BAD_ADDRESS, NULL));
    /* first check if consecutive */
    consecutive = 1;
    prev = list->addrs[0];
    if (list->nadd > 1)
	for (i=1; i < list->nadd; i++)
	    if (ntohl(list->addrs[i]) != ntohl(prev)+1) {
		consecutive = 0;
		break;
	    } else
		prev = list->addrs[i];
    for (i=0; i < lays; i++) {
	if (i < list->nadd) {
	    dstaddr[i] = list->addrs[i];
	} else {
	    dstaddr[i] = htonl(ntohl(dstaddr[i-1]) + 1);
	}
	if (!IN_CLASSD(ntohl(dstaddr[i])))
	    return (rlc_internal_set_error_(-1, ADDR_NOT_IN_CLASSD, NULL));
	NDEB (fprintf(stderr, "* have addr %lx\n", ntohl(dstaddr[i])));
    }
    return consecutive;
}

void
hton_addlist(addr_list_t *list)
{
  int i;
    for (i=0; i<list->nadd; i++)
	list->addrs[i] = htonl(list->addrs[i]);
}
 

int ui32diff(ui32 a, ui32 b)
{
    return((a-b));
}

long
diff_t(struct timeval b, struct timeval a)
{
    return (ui32)( 1000000*(b.tv_sec - a.tv_sec) +
	b.tv_usec - a.tv_usec) ;
}

long
timestamp_u()
{
    struct timeval t;

    (void)gettimeofday(&t, NULL);
    return((ui32)(t.tv_sec*1000000+t.tv_usec));
}

long
timestamp_m()
{
  struct timeval t;

    (void)gettimeofday(&t, NULL);
    return((ui32)(t.tv_sec*1000+t.tv_usec/1000));
}

long
timestamp_s()
{
  struct timeval t;

    (void)gettimeofday(&t, NULL);
    return((ui32)(t.tv_sec));
}

void
timestamp_t(struct timeval *t)
{
    (void)gettimeofday(t, NULL);
}


/**
 ** those are the stubs for events functions
 **/

/*++++
 Selects the next event to be executed or an IO descriptor ready
 for reading, and executes the associate function. If there are no
 events ready, it generally blocks, unless there are no timers
 pending and blocking = 0, in this case it just polls the IO descriptor
 and returns if none is ready for reading.

 Params:
 blocking : if = 0 doesn't blok in the select if
		there are no timers, it just polls.

 Returns:
	The id of the last event found or, if none,
	one of the following values ($< 0$):
	 - EV_DEFTIMEOUT  got the default timeout;
	 - EV_GOTSIGNAL  got a signal in select();
	 - EV_TERMINATE  no event selected because there are no
		timer pending and the IO descriptor list is empty, typically
		we must terminate the event loop;
	 - EV_NO_PENDING  no event selected because there are no timer
		pending and no IO descriptor was ready (only if blocking = 0).
*/
int
rlc_getevent(int blocking)
{
    return (rlc_getevent_(rlc_events, blocking));
}

int
rlc_poolchan(int *rid, void **rp)
{
    return (rlc_poolchan_(rlc_events, rid, rp));
}


void
rlc_set_default_timeout(unsigned long delta)
{
    return (rlc_set_default_timeout_(rlc_events, delta));
}
	
/*++++
 Adds an IO descriptor to the internal list of `observed' descriptors.
 The given function will be called when the descriptor is ready for
 reading.

 Params:
 	chan : IO descriptor being inserted.
 	f : function to be called.
 	p : function arguments.

 Returns:
	0 on success, < 0 on failure, > 0 if a warning
	is generated, use rlc_error_info() to get error
	or warning description. Possible error/warning
	causes are:
		- TOO_MANY_IOS (too many IO descriptors,
		Check the compile time defined constant MAX_CHAN).
*/
int
rlc_insertchan(int chan, int id, int (*f)(void *), void *p)
{
    return (rlc_insertchan_(rlc_events, chan, id, f, p));
}

/*++++
 Removes an IO descriptor from the internal list.

 Params:
 	chan : IO descriptor.
*/
void
rlc_deletechan(int chan)
{
    return (rlc_deletechan_(rlc_events, chan));
}


/*++++
 Returns the IO descriptor last found ready by rlc_getevent.

 Returns:
	the IO descriptor last found ready.
*/
int
rlc_showselected ()
{
    return (rlc_showselected_(rlc_events));
}

/*++++
 Arranges for a timer event to happen after delta usec.

 Params:
	id : id of the event to be added, will be returned
		by rlc_getevent after the event execution. It must be
		> 0 to avoid confusion with predefined
		return values.
	delta : relative time of event execution (in usec).
	f : function to be called.
	p : function arguments.
	real : if set, the event will occurr delta usec
		after the actual calling time. Otherwise, it is
		assumed that the function has been called by the
		scheduler, and the event is scheduled exactly delta
		usec after the scheduling function had to be
		executed. This allows not to accumulate errors.

 Returns:
	0 on success, $<$ 0 on failure. Possible error
	causes are: 
		- ALLOC_TIMIT_ERROR (Error allocating a timer
		descriptor)
*/
int
rlc_settimer(int id, unsigned long delta, CALLBACK f, void *p, int real)
{
    return (rlc_settimer_(rlc_events, id, delta, f, p, real));
}

/*++++
 Removes all the pending events with the given id,
 returns the number of events removed.

 Params:
	id : id of event(s) to be removed.
 Returns:
	the number of events removed.
*/
int
rlc_deletetimer(int id)
{
    return (rlc_deletetimer_(rlc_events, id));
}
